package be.iminds.ilabt.jfed.espec.model;

import be.iminds.ilabt.jfed.espec.parser.ESpecDefaults;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import io.dropwizard.jackson.Jackson;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.Collections;
import java.util.List;
import java.util.Map;

/**
 * AnsiblePlaybookSpec
 */
public class AnsiblePlaybookSpec extends UploadLikeWithLogSpec {
    private final int debug;
    @Nullable private final String extraVarsString;
    @Nullable private final ExtraVarsJsonFileSpec extraVars;

    private static final ObjectMapper MAPPER = Jackson.newObjectMapper();

    public AnsiblePlaybookSpec(@Nullable FileSource source,
                               @Nullable String path,
                               @Nullable String permissions,
                               @Nullable Integer debug,
                               @Nullable String log,
                               @Nullable ExtraVarsJsonFileSpec extraVars) {
        //nodes is special: always only on ansible node
        //                  value is ignored, but we put something anyway because null would mean ALL nodes and that's also not correct
        super(source, path, permissions == null ? "u=rwx" : permissions, Collections.singletonList("$ANSIBLE_NODE$"), log);

        this.debug = debug == null ? ESpecDefaults.ANSIBLE_DEBUG : debug;
        this.extraVars = extraVars;
        this.extraVarsString = null;
    }

    /**
     *
     * @param source
     * @param path
     * @param permissions
     * @param debug
     * @param log
     * @param extraVars in --extra-vars format
     */
    public AnsiblePlaybookSpec(@Nullable FileSource source,
                               @Nullable String path,
                               @Nullable String permissions,
                               @Nullable Integer debug,
                               @Nullable String log,
                               @Nullable String extraVars) {
        //nodes is special: always only on ansible node
        //                  value is ignored, but we put something anyway because null would mean ALL nodes and that's also not correct
        super(source, path, permissions == null ? "u=rwx" : permissions, Collections.singletonList("$ANSIBLE_NODE$"), log);

        this.debug = debug == null ? ESpecDefaults.ANSIBLE_DEBUG : debug;
        this.extraVars = null;
        this.extraVarsString = extraVars;
    }

    public AnsiblePlaybookSpec(@Nullable FileSource source,
                               @Nullable String path,
                               @Nullable String permissions,
                               @Nullable Integer debug,
                               @Nullable String log,
                               @Nonnull Map<String, Object> extraVars) {
        //nodes is special: always only on ansible node
        //                  value is ignored, but we put something anyway because null would mean ALL nodes and that's also not correct
        super(source, path, permissions == null ? "u=rwx" : permissions, Collections.singletonList("$ANSIBLE_NODE$"), log);

        this.debug = debug == null ? ESpecDefaults.ANSIBLE_DEBUG : debug;
        try {
            this.extraVars = extraVars.isEmpty() ?
                    null :
                    new ExtraVarsJsonFileSpec(new FileSource(FileSource.SourceType.DIRECT,
                            MAPPER.writeValueAsString(extraVars),
                            "extra-vars.json"));
        } catch (JsonProcessingException e) {
            throw new RuntimeException("Failed to convert extra vars into string", e);
        }
        this.extraVarsString = null;
    }


    public AnsiblePlaybookSpec(@Nullable FileSource source,
                               @Nullable String path,
                               @Nullable String permissions,
                               @Nullable Integer debug,
                               @Nullable String log) {
        this(source, path, permissions, debug, log, (String)null);
    }

    public boolean isDebug() {
        return debug > 0;
    }

    /**
     * @return number of v's to add to ansible arguments
     */
    public int getDebug() {
        return debug;
    }


    public boolean hasExtraVars() {
        return extraVarsString != null || extraVars != null;
    }
    public boolean hasExtraVarsString() {
        return extraVarsString != null;
    }
    public boolean hasExtraVarsJsonFileSpec() {
        return extraVars != null;
    }

    /**
     * @return  in --extra-vars format
     */
    @Nullable public String getExtraVarsString() {
        return extraVarsString;
    }
    @Nullable public ExtraVarsJsonFileSpec getExtraVarsJsonFileSpec() {
        return extraVars;
    }

    public static class ExtraVarsJsonFileSpec extends UploadLikeSpec {
        public ExtraVarsJsonFileSpec(@Nullable FileSource source,
                                     @Nullable String path,
                                     @Nullable String permissions,
                                     @Nullable List<String> nodes) {
            super(source, path, permissions == null ? "u=rwx" : permissions, nodes);
        }

        public ExtraVarsJsonFileSpec(@Nullable FileSource source) {
            this(source, null, null, null);
        }
    }
}
